==> lam_6.1.orig.tar.gz: lam-6.1.orig/share/mpi/reducescatter.c <==
/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	RBD/JRV
 *
 *	$Id: reducescatter.c,v 6.1 96/11/23 22:53:50 nevin Rel $
 *
 *	Function:	- reduce and scatter data to all processes
 *	Accepts:	- send buffer
 *			- receive buffer
 *			- counts of elements to recv
 *			- type of elements
 *			- operation to perform
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>
#include <terror.h>

int
MPI_Reduce_scatter(sbuf, rbuf, rcounts, dtype, op, comm)

void			*sbuf;
void			*rbuf;
int			*rcounts;
MPI_Datatype		dtype;
MPI_Op			op;
MPI_Comm		comm;

{
	int		i;
	int		size;			/* group size */
	int		rank;			/* my rank */
	int		count;			/* total count */
	int		err;			/* error code */
	int		*disps = 0;		/* displacement array */
	char		*buffer = 0;		/* reduce buffer */
	char		*origin = 0;		/* data origin in buffer */

	lam_initerr();
	lam_setfunc(BLKMPIREDUCESCATTER);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || LAM_IS_INTER(comm)) {
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_TYPE, 0)));
	}

	if (op == MPI_OP_NULL) {
		return(lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_OP, 0)));
	}

	if (rcounts == 0) {
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_COUNT, 0)));
	}

	LAM_TRACE(lam_tr_cffstart(BLKMPIREDUCESCATTER));
/*
 * Initialize reduce & scatterv info at the root (rank 0).
 */
	MPI_Comm_size(comm, &size);
	MPI_Comm_rank(comm, &rank);

	for (i = 0, count = 0; i < size; ++i) {
		if (rcounts[i] < 0) {
			return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
						lam_mkerr(MPI_ERR_COUNT, 0)));
		}
		count += rcounts[i];
	}

	if (rank == 0) {
		disps = (int *) malloc((unsigned) size * sizeof(int));
		if (disps == 0) {
			free((char *) disps);
			return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
					lam_mkerr(MPI_ERR_OTHER, errno)));
		}
		
		err = lam_dtbuffer(dtype, count, &buffer, &origin);
		if (err != MPI_SUCCESS) {
			free((char *) disps);
			return (lam_errfunc(comm, BLKMPIREDUCESCATTER, err));
		}

		disps[0] = 0;
		for (i = 0; i < (size - 1); ++i) {
			disps[i + 1] = disps[i] + rcounts[i];
		}
	}
/*
 * reduction
 */
	err = MPI_Reduce(sbuf, origin, count, dtype, op, 0, comm);
	if (err != MPI_SUCCESS) {
		if (disps) free((char *) disps);
		if (buffer) free(buffer);
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER, err));
	}
/*
 * scatter
 */
	err = MPI_Scatterv(origin, rcounts, disps, dtype,
				rbuf, rcounts[rank], dtype, 0, comm);

	if (disps) free((char *) disps);
	if (buffer) free(buffer);

	if (err != MPI_SUCCESS) {
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER, err));
	}

	LAM_TRACE(lam_tr_cffend(BLKMPIREDUCESCATTER,
					-1, comm, dtype, rcounts[rank]));

	lam_resetfunc(BLKMPIREDUCESCATTER);
	return(MPI_SUCCESS);
}

==> lam_6.3.2.orig.tar.gz: lam-6.3.2/share/mpi/reducescatter.c <==
/*
 *
 * Copyright 1998-1999, University of Notre Dame.
 * Authors: Jeffrey M. Squyres, Kinis L. Meyer with M. D. McNally 
 *          and Andrew Lumsdaine
 *
 * This file is part of the Notre Dame LAM implementation of MPI.
 *
 * You should have received a copy of the License Agreement for the
 * Notre Dame LAM implementation of MPI along with the software; see
 * the file LICENSE.  If not, contact Office of Research, University
 * of Notre Dame, Notre Dame, IN 46556.
 *
 * Permission to modify the code and to distribute modified code is
 * granted, provided the text of this NOTICE is retained, a notice that
 * the code was modified is included with the above COPYRIGHT NOTICE and
 * with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
 * file is distributed with the modified code.
 *
 * LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
 * By way of example, but not limitation, Licensor MAKES NO
 * REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
 * PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
 * OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
 * OR OTHER RIGHTS.  
 *
 * Additional copyrights may follow.
 *
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	RBD/JRV
 *
 *	$Id: reducescatter.c,v 6.3 1999/08/04 21:43:51 kmeyer1 Exp $
 *
 *	Function:	- reduce and scatter data to all processes
 *	Accepts:	- send buffer
 *			- receive buffer
 *			- counts of elements to recv
 *			- type of elements
 *			- operation to perform
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <lam_config.h>
#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>
#include <terror.h>

int
MPI_Reduce_scatter(sbuf, rbuf, rcounts, dtype, op, comm)

void			*sbuf;
void			*rbuf;
int			*rcounts;
MPI_Datatype		dtype;
MPI_Op			op;
MPI_Comm		comm;

{
	int		i;
	int		size;			/* group size */
	int		rank;			/* my rank */
	int		count;			/* total count */
	int		err;			/* error code */
	int		*disps = 0;		/* displacement array */
	char		*buffer = 0;		/* reduce buffer */
	char		*origin = 0;		/* data origin in buffer */

	lam_initerr();
	lam_setfunc(BLKMPIREDUCESCATTER);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || LAM_IS_INTER(comm)) {
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_TYPE, 0)));
	}

	if (op == MPI_OP_NULL) {
		return(lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_OP, 0)));
	}

	if (rcounts == 0) {
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
				lam_mkerr(MPI_ERR_COUNT, 0)));
	}

#if LAM_WANT_IMPI

	/* Remove this when IMPI collectives are implemented */

        if (LAM_IS_IMPI(comm)) {
	  return lam_err_comm(comm, MPI_ERR_COMM, 0, 
			      "Collectives not yet implemented on IMPI communicators");
	}
#endif 


	LAM_TRACE(lam_tr_cffstart(BLKMPIREDUCESCATTER));
/*
 * Initialize reduce & scatterv info at the root (rank 0).
 */
	MPI_Comm_size(comm, &size);
	MPI_Comm_rank(comm, &rank);

	for (i = 0, count = 0; i < size; ++i) {
		if (rcounts[i] < 0) {
			return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
						lam_mkerr(MPI_ERR_COUNT, 0)));
		}
		count += rcounts[i];
	}

	if (rank == 0) {
		disps = (int *) malloc((unsigned) size * sizeof(int));
		if (disps == 0) {
			free((char *) disps);
			return (lam_errfunc(comm, BLKMPIREDUCESCATTER,
					lam_mkerr(MPI_ERR_OTHER, errno)));
		}
		
		err = lam_dtbuffer(dtype, count, &buffer, &origin);
		if (err != MPI_SUCCESS) {
			free((char *) disps);
			return (lam_errfunc(comm, BLKMPIREDUCESCATTER, err));
		}

		disps[0] = 0;
		for (i = 0; i < (size - 1); ++i) {
			disps[i + 1] = disps[i] + rcounts[i];
		}
	}
/*
 * reduction
 */
	err = MPI_Reduce(sbuf, origin, count, dtype, op, 0, comm);
	if (err != MPI_SUCCESS) {
		if (disps) free((char *) disps);
		if (buffer) free(buffer);
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER, err));
	}
/*
 * scatter
 */
	err = MPI_Scatterv(origin, rcounts, disps, dtype,
				rbuf, rcounts[rank], dtype, 0, comm);

	if (disps) free((char *) disps);
	if (buffer) free(buffer);

	if (err != MPI_SUCCESS) {
		return (lam_errfunc(comm, BLKMPIREDUCESCATTER, err));
	}

	LAM_TRACE(lam_tr_cffend(BLKMPIREDUCESCATTER,
					-1, comm, dtype, rcounts[rank]));

	lam_resetfunc(BLKMPIREDUCESCATTER);
	return(MPI_SUCCESS);
}
